"""
API endpoint URLs for AskPablos Scrapy API.

This module contains internal API endpoints that are not meant to be configured by users.
"""

# API endpoint URL - not user configurable
API_URL = "http://10.10.10.178:7500/api/proxy/"
